/************************************************************************
 *  EXPRESSN.C                                                          *
 *                      Ausdrcke bewerten fr PINC                     *
 *                      ===========================                     *
 *                                                                      *
 *  21.06.93    Bodo Wenzel     V1.0, Erstellung                        *
 ************************************************************************/

/* === Einfgedateien ================================================= */

#include "ctype.h"
#include "stdio.h"
#include "string.h"

#include "pinc.h"
#include "preproc.h"
#include "compile.h"
#include "symbol.h"
#include "machine.h"

#define extern
#include "expressn.h"

/* === Funktionsprototypen ============================================ */

e_assign(int lval[]);           /* Zuweisung                            */
e_or(int lval[]);               /* bitxor { | bitxor }                  */
e_xor(int lval[]);              /* bitand { ^ bitand }                  */
e_and(int lval[]);              /* equality { & equality }              */
e_equal(int lval[]);            /* comparison { != == comparison }      */
e_comp(int lval[]);             /* shift { < > <= >= shift }            */
e_c_work(int op,int lval[]);    /* Zeigervergleich testen               */
e_c_ptr(int op);                /* Zeigervergleich Code                 */
e_shift(int lval[]);            /* addition { >> << addition }          */
e_add(int lval[]);              /* multiplicat { + - multiplicat }      */
e_mult(int lval[]);             /* unary { * / % unary }                */
e_unary(int lval[]);            /* ++ -- - * & primary ++ --            */
e_u_inc(int lval[]);            /* Wert oder Zeiger pre-erhhen         */
e_u_dec(int lval[]);            /* Wert oder Zeiger pre-erniedrigen     */
e_u_ref(int lval[]);            /* Adresse dereferenzieren              */
e_u_at(int lval[]);             /* Adresse referenzieren                */
e_u_id(int lval[]);             /* Wert oder Zeiger post-erhhen        */
e_u_di(int lval[]);             /* Wert oder Zeiger post-erniedrigen    */
e_primary(int lval[]);          /* Einfache Ausdrcke                   */
e_primitive(int lval[]);        /* einfache Typen                       */

e_const(int *val);              /* Konstante                            */
e_string(int *val);             /* Zeichenkette                         */

e_store(int lval[]);            /* Speichern eines Wertes               */
e_rvalue(int lval[]);           /* Laden eines Wertes                   */

/* === Funktionen ===================================================== */

expinit()                       /* Initialisierung                      */
/*-----*/
  {
  litptr=0;                     /* bisher keine Konstanten              */
  }

expression()                    /* Ausdruck compilieren                 */
/*--------*/
  {
  int   lval[2];

  if (e_assign(lval))           /* wenn Zuweisung,                      */
    e_rvalue(lval);             /* Wert speichern                       */
  }

e_assign(int lval[])            /* Zuweisung                            */
/*----------------*/
  {
  int   k,lval2[2];

  k=e_or(lval);
  if (matchx("="))
    {
    if (k==0)
      {
      erlvalue();
      lval[1]=S_TY_INT;
      }
    if (lval[1])
      mpush();
    if (e_assign(lval2))
      e_rvalue(lval2);
    e_store(lval);
    return(0);
    }
  else
    return(k);
  }

/* --- Funktionen fr Operatoren -------------------------------------- */

e_or(int lval[])                /* bitxor { | bitxor }                  */
/*------------*/
  {
  int   k,lval2[2];

  k=e_xor(lval);
  if (ppactch()!='|')
    return(k);
  if (k)
    e_rvalue(lval);
  while (1)
    {
    if (matchx("|"))
      {
      mpush();
      if (e_xor(lval2))
        e_rvalue(lval2);
      mpop();
      moutcmd(T_OR);
      }
    else
      return(0);
    }
  }

e_xor(int lval[])               /* bitand { ^ bitand }                  */
/*-------------*/
  {
  int   k,lval2[2];

  k=e_and(lval);
  if (ppactch()!='^')
    return(k);
  if (k)
    e_rvalue(lval);
  while (1)
    {
    if (matchx("^"))
      {
      mpush();
      if (e_and(lval2))
        e_rvalue(lval2);
      mpop();
      moutcmd(T_XOR);
      }
    else
      return(0);
    }
  }

e_and(int lval[])               /* equality { & equality }              */
/*-------------*/
  {
  int   k,lval2[2];

  k=e_equal(lval);
  if (ppactch()!='&')
    return(k);
  if (k)
    e_rvalue(lval);
  while (1)
    {
    if (matchx("&"))
      {
      mpush();
      if (e_equal(lval2))
        e_rvalue(lval2);
      mpop();
      moutcmd(T_AND);
      }
    else
      return(0);
    }
  }

e_equal(int lval[])             /* comparison { != == comparison }      */
/*---------------*/
  {
  int   k,lval2[2];

  k=e_comp(lval);
  if ((ppactch()!='=' & ppch0()!='!') | ppch1()!='=')
    return(k);
  if (k)
    e_rvalue(lval);
  while (1)
    {
    if (matchx("=="))
      {
      mpush();
      if (e_comp(lval2))
        e_rvalue(lval2);
      mpop();
      moutcmd(T_EQ);
      }
    else if (matchx("!="))
      {
      mpush();
      if (e_comp(lval2))
        e_rvalue(lval2);
      mpop();
      moutcmd(T_NE);
      }
    else
      return(0);
    }
  }

e_comp(int lval[])              /* shift { < > <= >= shift }            */
/*--------------*/
  {
  int   k;

  k=e_shift(lval);
  if (ppactch()!='<' & ppch0()!='>')
    return(k);
  if (ppch0()=='>' & ppch1()=='>')
    return(k);
  if (ppch0()=='<' & ppch1()=='<')
    return(k);
  if (k)
    e_rvalue(lval);
  while (1)
    {
    if (matchx("<="))
      {
      if (e_c_work(1,lval))
        continue;
      moutcmd(T_SLE);
      }
    else if (matchx(">="))
      {
      if (e_c_work(2,lval))
        continue;
      moutcmd(T_SGE);
      }
    else if (ppactch()=='<')
      {
      if (ppch1()=='<')
        return(0);
      ppgetch();
      if (e_c_work(3,lval))
        continue;
      moutcmd(T_SLT);
      }
    else if (ppactch()=='>')
      {
      if (ppch1()=='>')
        return(0);
      ppgetch();
      if (e_c_work(4,lval))
        continue;
      moutcmd(T_SGT);
      }
    else
      return(0);
    }
  }

e_c_work(int op,int lval[])     /* Zeigervergleich testen               */
/*-----------------------*/
  {
  int   lval2[2];
  char  *s;

  mpush();
  if (e_shift(lval2))
    e_rvalue(lval2);
  mpop();
  if (s=lval[0])
    if (s[SYM_ID]==S_ID_PTR)
      {
      e_c_ptr(op);
      return(1);
      }
  if (s=lval2[0])
    if (s[SYM_ID]==S_ID_PTR)
      {
      e_c_ptr(op);
      return(1);
      }
  return(0);
  }

e_c_ptr(int op)                 /* Zeigervergleich Code                 */
/*-----------*/
  {
  switch (op)
    {
    case 1:
      moutcmd(T_ULE);
      break;
    case 2:
      moutcmd(T_UGE);
      break;
    case 3:
      moutcmd(T_ULT);
      break;
    default:
      moutcmd(T_UGT);
      break;
    }
  }

e_shift(int lval[])             /* addition { >> << addition }          */
/*---------------*/
  {
  int   k,lval2[2];

  k=e_add(lval);
  if ((ppactch()!='>' | ppch1()!='>') &
      (ppch0()!='<' | ppch1()!='<'))
    return(k);
  if (k)
    e_rvalue(lval);
  while (1)
    {
    if (matchx(">>"))
      {
      mpush();
      if (e_add(lval2))
        e_rvalue(lval2);
      mpop();
      moutcmd(T_SHR);
      }
    else if (matchx("<<"))
      {
      mpush();
      if (e_add(lval2))
        e_rvalue(lval2);
      mpop();
      moutcmd(T_SHL);
      }
    else
      return(0);
    }
  }

e_add(int lval[])               /* multiplicat { + - multiplicat }      */
/*-------------*/
  {
  int   k,lval2[2];
  char  *s;

  k=e_mult(lval);
  if (ppactch()!='+' & ppch0()!='-')
    return(k);
  if (k)
    e_rvalue(lval);
  while (1)
    {
    if (matchx("+"))
      {
      mpush();
      if (e_mult(lval2))
        e_rvalue(lval2);
      if (s=lval[0])
        if (s[SYM_ID]==S_ID_PTR &
            s[SYM_TYPE]==S_TY_INT)
          moutcmd(T_CPTR);
      mpop();
      moutcmd(T_ADD);
      }
    else if (matchx("-"))
      {
      mpush();
      if (e_mult(lval2))
        e_rvalue(lval2);
      if (s=lval[0])
        if (s[SYM_ID]==S_ID_PTR &
            s[SYM_TYPE]==S_TY_INT)
          moutcmd(T_CPTR);
      mpop();
      moutcmd(T_SUB);
      }
    else
      return(0);
    }
  }

e_mult(int lval[])              /* unary { * / % unary }                */
/*--------------*/
  {
  int   k,lval2[2];

  k=e_unary(lval);
  if (ppactch()!='*' & ppch0()!='/' & ppch0()!='%')
    return(k);
  if (k)
    e_rvalue(lval);
  while (1)
    {
    if (matchx("*"))
      {
      mpush();
      if (e_unary(lval2))
        e_rvalue(lval2);
      mpop();
      moutcmd(T_MUL);
      }
    else if (matchx("/"))
      {
      mpush();
      if (e_unary(lval2))
        e_rvalue(lval2);
      mpop();
      moutcmd(T_DIV);
      }
    else if (matchx("%"))
      {
      mpush();
      if (e_unary(lval2))
        e_rvalue(lval2);
      mpop();
      moutcmd(T_MOD);
      }
    else
      return(0);
    }
  }

/* --- die untersten Stufen ------------------------------------------- */

e_unary(int lval[])             /* ++ -- - * & primary ++ --            */
/*---------------*/
  {
  int   k;

  if (matchx("++"))
    {
    if (e_unary(lval))
      e_u_inc(lval);
    else
      erlvalue();
    return(0);
    }
  else if (matchx("--"))
    {
    if (e_unary(lval))
      e_u_dec(lval);
    else
      erlvalue();
    return(0);
    }
  else if (matchx("-"))
    {
    if (e_unary(lval))
      e_rvalue(lval);
    moutcmd(T_NEG);
    return(0);
    }
  else if (matchx("*"))
    {
    e_u_ref(lval);
    return(1);
    }
  else if (matchx("&"))
    {
    if (e_unary(lval)==0)
      error("Illegale Adresse");
    else if (lval[1]==S_TY_NONE)
      e_u_at(lval);
    return(0);
    }
  else
    {
    k=e_primary(lval);
    if (matchx("++"))
      {
      if (k)
        e_u_id(lval);
      else
        erlvalue();
      return(0);
      }
    else if (matchx("--"))
      {
      if (k)
        e_u_di(lval);
      else
        erlvalue();
      return(0);
      }
    else
      return(k);
    }
  }

e_u_inc(int lval[])             /* Wert oder Zeiger pre-erhhen         */
/*---------------*/
  {
  char  *s;

  if (lval[1])
    mpush();
  e_rvalue(lval);
  s=lval[0];
  if (s[SYM_ID]==S_ID_PTR )
    {
    if (s[SYM_TYPE]==S_TY_INT)
      moutcmd(T_INCI);
    else
      moutcmd(T_INC);
    }
  else
    moutcmd(T_INC);
  e_store(lval);
  }

e_u_dec(int lval[])             /* Wert oder Zeiger pre-erniedrigen     */
/*---------------*/
  {
  char  *s;

  if (lval[1])
    mpush();
  e_rvalue(lval);
  s=lval[0];
  if (s[SYM_ID]==S_ID_PTR )
    {
    if (s[SYM_TYPE]==S_TY_INT)
      moutcmd(T_DECI);
    else
      moutcmd(T_DEC);
    }
  else
    moutcmd(T_DEC);
  e_store(lval);
  }

e_u_ref(int lval[])             /* Adresse dereferenzieren              */
/*---------------*/
  {
  int   k;
  char  *s;

  k=e_unary(lval);
  if (k)
    e_rvalue(lval);
  lval[1]=S_TY_INT;
  if (s=lval[0])
    lval[1]=s[SYM_TYPE];
  lval[0]=NULL;
  }

e_u_at(int lval[])              /* Adresse referenzieren                */
/*--------------*/
  {
  char  *s;

  mgetgadr(s=lval[0]);
  lval[1]=s[SYM_TYPE];
  }

e_u_id(int lval[])              /* Wert oder Zeiger post-erhhen        */
/*--------------*/
  {
  char  *s;

  if (lval[1])
    mpush();
  e_rvalue(lval);
  s=lval[0];
  if (s[SYM_ID]==S_ID_PTR )
    {
    if (s[SYM_TYPE]==S_TY_INT)
      moutcmd(T_INCI);
    else
      moutcmd(T_INC);
    }
  else
    moutcmd(T_INC);
  e_store(lval);
  if (s[SYM_ID]==S_ID_PTR )
    {
    if (s[SYM_TYPE]==S_TY_INT)
      moutcmd(T_DECI);
    else
      moutcmd(T_DEC);
    }
  else
    moutcmd(T_DEC);
  }

e_u_di(int lval[])              /* Wert oder Zeiger post-erniedrigen    */
/*--------------*/
  {
  char  *s;

  if (lval[1])
    mpush();
  e_rvalue(lval);
  s=lval[0];
  if (s[SYM_ID]==S_ID_PTR )
    {
    if (s[SYM_TYPE]==S_TY_INT)
      moutcmd(T_DECI);
    else
      moutcmd(T_DEC);
    }
  else
    moutcmd(T_DEC);
  e_store(lval);
  if (s[SYM_ID]==S_ID_PTR )
    {
    if (s[SYM_TYPE]==S_TY_INT)
      moutcmd(T_INCI);
    else
      moutcmd(T_INC);
    }
  else
    moutcmd(T_INC);
  }

e_primary(int lval[])           /* Einfache Ausdrcke                   */
/*-----------------*/
  {
  int   k;
  char  *s;

  k=e_primitive(lval);
  s=lval[0];
  if (ppactch()=='[' | ppch0()=='(')
    while (1)                   /* Fehler, wenn mehrfache Indizierung!  */
      {
      if (matchx("["))
        {
        if (s==NULL)
          {
          error("Illegaler Index");
          skipjunk();
          ccneedchar("]");
          return(0);
          }
        else if (s[SYM_ID]==S_ID_PTR)
          e_rvalue(lval);
        else if (s[SYM_ID]!=S_ID_ARR)
          error("Illegaler Index");
        mpush();
        expression();
        ccneedchar("]");
        if (s[SYM_TYPE]==S_TY_INT)
          moutcmd(T_CPTR);
        mpop();
        moutcmd(T_ADD);
        lval[1]=s[SYM_TYPE];
        k=1;
        }
      else if (matchx("("))
        {
        if (s==NULL)
          cccall(NULL);
        else if (s[SYM_ID]!=S_ID_FNC)
          {
          e_rvalue(lval);
          cccall(NULL);
          }
        else
          cccall(s);
        k=0;
        lval[0]=NULL;
        }
      else
        return(k);
      }
  if (s==NULL)
    return(k);
  if (s[SYM_ID]==S_ID_FNC)
    {
    mgetgadr(s);
    return(0);
    }
  return(k);
  }

e_primitive(int lval[])         /* einfache Typen                       */
/*-------------------*/
  {
  char  *s,name[NAMESIZE];
  int   val,k,*iptr;

  if (matchx("("))
    {
    k=e_assign(lval);
    ccneedchar(")");
    return(k);
    }
  if (symname(name))
    {
    if (s=symfind(name))
      {
      if (s>=symgptr)
        {
        mgetladr(s);
        lval[0]=s;
        lval[1]=s[SYM_TYPE];
        if (s[SYM_ID]==S_ID_PTR)
          lval[1]=S_TY_INT;
        if (s[SYM_ID]==S_ID_ARR)
          return(0);
        else
          return(1);
        }
      else
        {
        iptr=s+SYM_OFFSET;
        if (*iptr==S_OF_EXT)
          *iptr=S_OF_REF;
        lval[0]=s;
        lval[1]=S_TY_NONE;
        if (s[SYM_ID]==S_ID_FNC)
          return(0);
        if (s[SYM_ID]!=S_ID_ARR)
          return(1);
        mgetgadr(s);
        lval[1]=s[SYM_TYPE];
        return(0);
        }
      }
    error("Unbekanntes Symbol");
    moutcmdv(T_LIMM,0);
    lval[0]=0;
    lval[1]=S_TY_NONE;
    return(0);
    }
  if (e_const(&val)==0)
    {
    error("Ungltiger Ausdruck");
    moutcmdv(T_LIMM,0);
    skipjunk();
    }
  lval[0]=0;
  lval[1]=S_TY_NONE;
  return(0);
  }

e_const(int *val)               /* Konstante                            */
/*-------------*/
  {
  if (enumber(val))
    moutcmdv(T_LIMM,*val);
  else if (echar(val))
    moutcmdv(T_LIMM,*val);
  else if (e_string(val))
    mgetdadr(*val);
  else
    return(0);
  return(1);
  }

enumber(int *val)               /* Zahl in Ziffern                      */
/*-------------*/
  {
  int   k,minus;

  k=0;
  if (matchx("+"))
    minus=0;
  else
    minus=matchx("-");
  if (isdigit(ppactch())==0)
    return(0);
  if (ppch0()=='0' & ppch1()=='x')
    {
    ppgetch();
    ppgetch();
    while (1)
      if (isdigit(ppch0()))
        k=k*16+ppgetch()-'0';
      else if (toupper(ppch0())>='A' & toupper(ppch0())<='F')
        k=k*16+toupper(ppgetch())-'A'+10;
      else
        break;
    }
  else
    do
      k=k*10+ppgetch()-'0';
    while (isdigit(ppch0()));
  if (minus)
    *val=-k;
  else
    *val=k;
  return(1);
  }

echar(int *val)                 /* Einzelzeichen                        */
/*-----------*/
  {
  int   k;
  char  c;

  k=0;
  if (matchx("'")==0)
    return(0);
  while ((c=ppgetch())!='\'')
    {
    if (c=='\\')
      switch (c=ppgetch())
        {
        case '0':
          c=0x00;
          break;
        case 'b':
          c=0x08;
          break;
        case 't':
          c=0x09;
          break;
        case 'n':
          c=0x0A;
          break;
        case 'f':
          c=0x0C;
          break;
        case 'r':
          c=0x0D;
          break;
        default:
          ;
        }
    k=(k<<8)+c;
    }
  *val=k;
  return(1);
  }

e_string(int *val)              /* Zeichenkette                         */
/*--------------*/
  {
  char  c;

  if (matchx("\"")==0)
    return(0);
  *val=litptr;
  while ((c=ppgetch())!='"')
    {
    if (c=='\\')
      switch (c=ppgetch())
        {
        case '0':
          c=0x00;
          break;
        case 'b':
          c=0x08;
          break;
        case 't':
          c=0x09;
          break;
        case 'n':
          c=0x0A;
          break;
        case 'f':
          c=0x0C;
          break;
        case 'r':
          c=0x0D;
          break;
        default:
          ;
        }
    if (litptr<LITEND)
      lit[litptr++]=c;
    }
  if (litptr>=LITEND)
    error("Konstantenspeicher voll");
  else
    lit[litptr++]='\0';
  return(1);
  }

/* --- Speichern und Lesen eines Wertes ------------------------------- */

e_store(int lval[])             /* Speichern eines Wertes               */
/*---------------*/
  {
  if (lval[1]==S_TY_NONE)
    mputgbl(lval[0]);
  else
    mputind(lval[1]);
  }

e_rvalue(int lval[])            /* Laden eines Wertes                   */
/*----------------*/
  {
  if (lval[0]!=NULL & lval[1]==S_TY_NONE)
    mgetgbl(lval[0]);
  else
    mgetind(lval[1]);
  }

/* === Ende =========================================================== */
