/************************************************************************
 *  PINC.C                                                              *
 *              PINC ist nicht C - ein Compiler fr alle                *
 *              ========================================                *
 *                                                                      *
 * Original von Ron Cain und im Dr. Dobbs's Journal, #45, Volume V,     *
 * Number 5 als SMALL-C beschrieben.                                    *
 *  xx.03.93    Bodo Wenzel     V0.0, Quelle von SMALL-C V1.1N          *
 *  28.08.93    Bodo Wenzel     V1.0, viele neue Features, Code fr     *
 *                              fast beliebige Prozessoren              *
 ************************************************************************/

/* === Prolog ========================================================= */

/* Es begann alles damit, da ich von einem Benutzer meines Assemblers
   (ASMxxx auf Atari ST) die Anfrage bekam, ob ich denn einen C-Compiler
   fr den Mikrocontroller 8051 kennen wrde. "Natrlich" kannte ich
   einen, aber der lief nur auf DOS-Kompatiblen. Und da ich schon immer
   wissen wollte, wie man mglichst einfach einen Compiler schreibt,
   wuchs der Wunsch nach einem eigenen, wenn mglich CPU-unabhngigen,
   kleinen Compiler.

   Nicht viel spter las ich durch Zufall etwas ber SMALL-C, das im
   ausgehenden CP/M- und angehenden MS-DOS-Zeitalter recht bekannt war.
   Mit Hilfe einer kleinen Umfrage unter meinen Informatik-Bekannten
   bekam ich dann den Quelltext und die fertige Version fr PC-DOS in
   die Hnde. Sie funktionierte recht brauchbar (vor allem, wenn man
   sich auf Mikrocontroller beschrnken wollte) und erzeugte Quelltexte
   fr Assembler... Gegen Ende meiner Bearbeitung las ich auch noch von
   der wohl neuesten Version V2.2, die ich aber nicht zu fassen bekam.
   Sie hat hnliche Erweiterungen wie PINC, arbeitet aber nicht so
   flexibel.

   Nun nahm ich den Quelltext von SMALL-C, bearbeitete ihn ein bichen
   und baute auch Laufzeitsysteme fr die 680x0-CPU des Ataris unter TOS
   und fr die 80x86-CPU der Kompatiblen unter DOS. Dazu kreierte ich
   einige Codetabellen fr verschiedene Prozessoren.

   Jetzt fehlte mir nur noch ein Name. Zu der Zeit war Atari gerade
   dabei, das Betriebssystem auf Multitasking umzustellen und daher war
   der Name des neuen Kernels in aller Munde: MiNT. Mir gefielen diese
   "selbstbeinhaltenden" Abkrzungen schon immer und so fiel die Wahl
   auf PINC: "Pinc Ist Nicht C" (womit ich auf die starken Beschrnkungen
   gegenber einem richtigen C-Compiler anspiele).

   Wie der Compiler funktioniert und was alles in der Systembibliothek
   enthalten ist, steht in der Dokumentation. Hier hat das nichts zu
   suchen, wer schaut schon in den Quelltext...
*/

/* === Einfgedateien ================================================= */

#include "ctype.h"
#include "stdio.h"
#include "stdlib.h"
#include "string.h"

#include "preproc.h"
#include "compile.h"
#include "symbol.h"
#include "expressn.h"
#include "machine.h"

#define extern
#include "pinc.h"

#include "kbhit.c"
/* Diese Datei mu fr jedes System neu geschrieben werden...           */

/* === Globale Variablen ============================================== */

char    *run;                   /* Zeiger auf Laufanzeige               */
int     errfile,                /* Fehlerdatei                          */
        errcnt;                 /* Anzahl Fehler                        */
char    eline[OLINESIZE];       /* Puffer fr Fehlermeldungen           */

/* === Funktionsprototypen ============================================ */

p_usage();                      /* Aufrufhilfe                          */
p_arg(char *s);                 /* Nchstes Argument lesen              */
p_abort(int n);                 /* Abbruch                              */
er_sum();                       /* Fehler-Zusammenfassung               */

/* === Funktionen ===================================================== */

main(char *arg)
/*-----------*/
  {
  char  *s,tabname[ILINESIZE];

  ininit();
  errfile=stderr;
  outfile=NULL;
  verbose=0;                    /* Variablen vorbesetzen                */

  s=p_arg(arg);                 /* Name der Quelldatei                  */
  if (ilineptr==iline)
    p_usage();
  if (inopen(iline)==0)
    errfatal("Kann Quelldatei nicht ffnen",RET_OPEN);

  s=p_arg(s);                   /* Name der bersetzungstabelle         */
  if (ilineptr==iline)
    p_usage();
  strcpy(tabname,iline);

  while (1)
    {
    s=p_arg(s);                 /* nchstes Argument                    */
    if (ilineptr==iline)
      break;
    if (iline[0]!='-')
      p_usage();
    switch (toupper(iline[1]))  /* je nach Option...                    */
      {
      case 'O':                 /* Ausgabedatei                         */
        if (outfile==NULL)
          if ((outfile=fopen(iline+2,"w"))==NULL)
            errfatal("Kann Ausgabedatei nicht ffnen",RET_OPEN);
        break;
      case 'E':                 /* Fehlerdatei                          */
        if (errfile==stderr)
          if ((errfile=fopen(iline+2,"w"))==NULL)
            {
            errfile=stderr;
            errfatal("Kann Fehlerdatei nicht ffnen",RET_OPEN);
            }
        break;
      case 'V':                 /* ausfhrliche Lauf-Ausgabe            */
        verbose=1;
        break;
      default:
        p_usage();
      }
    }

  title="PINC C-Compiler V1.0, Copyright 1993 Bodo Wenzel";
  if (verbose)
    {
    puts(title);
    putchar('\r');
    putchar('\n');
    run="/-\\|";
    }

  syminit();
  expinit();
  mreadtab(tabname);
  preinit();                    /* Initialisierungen                    */

  errcnt=0;
  mheader(arg);
  moutcmd(T_BCODE);
  compile();                    /* compilieren !                        */
  mextfuncs();
  moutcmd(T_ECODE);
  moutcmd(T_BDATA);
  mdataout();
  moutcmd(T_EDATA);
  moutcmd(T_BBSS);
  mbssout();
  moutcmd(T_EBSS);
  mtrailer();
  er_sum();                     /* Fehler melden                        */

  if (outfile!=NULL)
    fclose(outfile);
  if (errfile!=stderr)
    fclose(errfile);
  return(RET_NONE);
  }

p_usage()                       /* Aufrufhilfe                          */
/*-----*/
  {
  fputs("Aufruf: PINC quelle tabelle [-Oausgabe] [-Efehler] [-V]\r\n",
        stderr);
  fputs("quelle     C-Quelltext in \"quelle\"\r\n",stderr);
  fputs("tabelle    bersetzungstabelle in \"tabelle\"\r\n",stderr);
  fputs("-Oausgabe  Erzeugt Assemblertext in \"ausgabe\"\r\n",stderr);
  fputs("-Efehler   Fehlermeldungen in \"fehler\"\r\n",stderr);
  fputs("-V         PINC als Plaudertasche\r\n",stderr);
  exit(RET_ARG);
  }

p_arg(char *s)                  /* Nchstes Argument lesen              */
/*----------*/
  {
  --s;
  while (isspace(*++s));        /* Leerzeichen berspringen             */
  ilineptr=iline;
  while (isspace(*s)==0 & *s!='\0')
    *ilineptr++=*s++;
  *ilineptr='\0';               /* nchstes Wort kopieren               */
  return(s);
  }

testabort()                     /* Bei Bedarf abbrechen                 */
/*-------*/
  {
  if (kbhit())                  /* einige Betriebssysteme fangen den    */
    if (getchar()==0x03)        /* Abbruch selbst ab!                   */
      p_abort(RET_ABORT);

  if (verbose)                  /* auf Wunsch Laufanzeige               */
    {
    putchar(*run);
    putchar('\b');
    if (*++run=='\0')
      run=run-4;
    }
  }

p_abort(int n)                  /* Abbruch                              */
/*----------*/
  {
  puts("Compilation abgebrochen.\r\n");
  exit(n);
  }

/* === Fehlermeldungen ================================================ */

error(char *s)                  /* Fehler melden                        */
/*----------*/
  {
  if (in_i>=0)
    sprintf(eline,"Zeile %d",fileline[in_i]);
  else
    strcpy(eline,"Fehler");
  fputs(eline,errfile);
  if (cfunc)
    {
    sprintf(eline,", %s()",cfunc+SYM_NAME);
    fputs(eline,errfile);
    sprintf(eline," + %d",funcline);
    fputs(eline,errfile);
    }
  sprintf(eline,": %s!\r\n",s);
  fputs(eline,errfile);
  fputs(pline,errfile);
  fputc('\r',errfile);
  fputc('\n',errfile);
  ++errcnt;
  }

erpointer()                     /* Mehrdimensionalitt melden           */
/*-------*/
  {
  error("Vektoren von Zeigern nicht implementiert");
  }

ername()                        /* Fehler im Symboltext melden          */
/*----*/
  {
  error("Fehlerhafter Symbolname");
  skipjunk();
  }

ermulti(char *s)                /* Doppelte Definition melden           */
/*------------*/
  {
  error("Bereits definiert");
  moutcmds(T_COMM,s);
  }

erlvalue()                      /* "Lvalue"-Fehler melden               */
/*------*/
  {
  error("Links von \"=\" mu ein \"lvalue\" sein");
  }

er_sum()                        /* Fehler-Zusammenfassung               */
/*----*/
  {
  sprintf(eline,"Es gab %d Fehler beim Compilieren.",errcnt);
  if (verbose)
    {
    puts(eline);
    putchar('\r');
    putchar('\n');
    }
  if (errcnt)
    {
    fputs(eline,errfile);
    fputc('\r',errfile);
    if (fputc('\n',errfile)==EOF)
      errfatal("Schreibfehler auf Fehlerdatei",RET_WRITE);
    }
  moutcmds(T_COMM,eline);
  }

errfatal(char *s,int n)         /* Schweren Fehler melden               */
/*-------------------*/
  {
  sprintf(eline,"Fataler Fehler: %s!\r\n",s);
  fputs(eline,stderr);
  p_abort(n);
  }

/* === Ende =========================================================== */
