/************************************************************************
 *  PREPROC.C                                                           *
 *                      Der Preprozessor fr PINC                       *
 *                      =========================                       *
 *                                                                      *
 *  26.06.93    Bodo Wenzel     V1.0, Erstellung                        *
 ************************************************************************/

/* === Einfgedateien ================================================= */

#include "ctype.h"
#include "stdio.h"
#include "string.h"

#include "pinc.h"
#include "symbol.h"
#include "machine.h"

#define extern
#include "preproc.h"

/* === Konstanten ===================================================== */

#define MACMAX          2000
#define MACEND          1999    /* MACMAX-1                             */

/* === Globale Variablen ============================================== */

char    mac[MACMAX],            /* Makro-Tabelle                        */
        *macptr;                /* Index auf freien Platz               */

/* === Funktionsprototypen ============================================ */

in_line();                      /* eine Zeile Text lesen                */

p_expand();                     /* Nchstes Token lesen                 */
p_addch(char c);                /* Zeichen hinzufgen                   */
p_string();                     /* Zeichenkette                         */
p_comment();                    /* Kommentar                            */
p_number();                     /* Wort mit Ziffer am Anfang            */
p_word();                       /* Wort mit Buchstaben am Anfang        */
p_command();                    /* Kommando ausfhren                   */
p_include();                    /* Einfgedatei ffnen                  */
p_asm();                        /* Direkte Ausgabe an Zieldatei         */
p_match(char *s);               /* Wortvergleich                        */

mac_add();                      /* Makro hinzufgen                     */
mac_put(char c);                /* Makrozeichen hinzufgen              */
mac_find(char *name);           /* Makro suchen                         */

/* === Funktionen ===================================================== */

/* --- Quelldateien --------------------------------------------------- */

ininit()                        /* Initialisierung                      */
/*----*/
  {
  ineof=0;                      /* bisher kein Quellenende              */
  in_i=-1;                      /* keine Datei offen                    */
  }

inopen(char *name)              /* Nchste Datei ffnen                 */
/*--------------*/
  {
  if ((infile[in_i+1]=fopen(name,"r"))==NULL)   /* Fehler ?             */
    return(0);
  ++in_i;
  fileline[in_i]=0;             /* Zeilenzhler setzen                  */
  return(1);
  }

in_line()                       /* eine Zeile Text lesen                */
/*-----*/
  {
  char  c;

  testabort();                  /* bei Bedarf abbrechen                 */

  if (in_i<0)                   /* alle Dateien geschlossen ?           */
    {
    ineof=1;
    return;
    }
  ++funcline;
  ++fileline[in_i];             /* Zeilenzhler erhhen                 */

  pline[0]='\0';
  if (fgets(pline,ILINEEND,infile[in_i])==NULL) /* Zeile lesen          */
    fclose(infile[in_i--]);
  for (plineptr=pline; *plineptr; ++plineptr)
    if ((c=*plineptr)=='\r' | c=='\n' | c==0x1A)
      {
      *plineptr='\0';           /* Zeilenende lschen                   */
      break;
      }
  }

/* --- Zeichen aus dem Preprozessor ----------------------------------- */

ppactch()                       /* aktuelles Zeichen mit Holen          */
/*-----*/
  {
  while (*ilineptr=='\0')       /* bis es effektiv etwas gibt...        */
    {
    if (ineof)
      return(0);
    preprocess();               /* Preprozessor aufrufen                */
    }
  return(*ilineptr);
  }

ppch0()                         /* aktuelles Zeichen                    */
/*---*/
  {
  return(*ilineptr);
  }

ppch1()                         /* nchstes Zeichen                     */
/*---*/
  {
  if (ppch0())
    return(ilineptr[1]);
  else
    return('\0');
  }

ppgetch()                       /* aktuelles Zeichen und weiter         */
/*-----*/
  {
  char  c;

  if (c=ppactch())
    ++ilineptr;
  return(c);
  }

/* --- Der Preprozessor ----------------------------------------------- */

preinit()                       /* Initialisierung                      */
/*-----*/
  {
  pline[0]='\0';
  plineptr=pline;
  iline[0]='\0';
  ilineptr=iline;
  macptr=mac;                   /* Zeiger zurcksetzen                  */
  }

preprocess()                    /* Preprozessor                         */
/*--------*/
  {
  while (*plineptr=='\0')       /* wenn es nicht mehr gibt...           */
    {
    if (ineof)
      return;
    in_line();                  /* eine Zeile Quelltext holen           */

    moutcmds(T_COMM,pline);
    plineptr=pline;
    while (isspace(*plineptr))
      ++plineptr;
    if (*plineptr++=='#')       /* bei Bedarf Preprozessor-Kommando     */
      p_command();
    plineptr=pline;
    }

  ilineptr=iline;
  *ilineptr='\0';
  p_expand();                   /* nchstes Token expandieren           */
  if (ilineptr>=iline+ILINEEND)
    error("Zeile zu lang");
  ilineptr=iline;               /* Zeiger auf den Anfang                */
  }

p_expand()                      /* Nchstes Token lesen                 */
/*------*/
  {
  int   first;

  first=1;                      /* noch nichts bernommen               */
  while (*plineptr)
    {
    if ((*plineptr=='\'') | (*plineptr=='"'))
      {
      if (first)
        p_string();
      break;
      }
    if ((*plineptr=='/') & (plineptr[1]=='*'))
      {
      if (first)
        p_comment();
      break;
      }
    if (isdigit(*plineptr))
      {
      if (first)
        p_number();
      break;
      }
    if (alpha(*plineptr))
      {
      if (first)
        p_word();
      break;
      }
    if (isspace(*plineptr))
      {
      if (first)
        plineptr++;
      else
        break;
      }
    else
      {
      first=0;
      p_addch(*plineptr++);
      }
    }
  *ilineptr='\0';
  }

p_addch(char c)                 /* Zeichen hinzufgen                   */
/*-----------*/
  {
  if (ilineptr<iline+ILINEEND)
    *ilineptr++=c;
  return(c);
  }

p_string()                      /* Zeichenkette                         */
/*------*/
  {
  char  delim;

  p_addch(delim=*plineptr++);
  while (*plineptr!=delim)
    {
    if (*plineptr=='\0')
      {
      error("Fehlendes Anfhrungszeichen");
      break;
      }
    if (*plineptr=='\\' & plineptr[1]!=0)
      p_addch(*plineptr++);
    p_addch(*plineptr++);
    }
  ++plineptr;
  p_addch(delim);
  }

p_comment()                     /* Kommentar                            */
/*-------*/
  {
  ++plineptr;
  ++plineptr;
  while ((*plineptr!='*') | (plineptr[1]!='/'))
    {
    if (*plineptr=='\0')
      {
      if (ineof)
        return;
      in_line();                /* nchste Zeile ohne bersetzung       */
      plineptr=pline;
      moutcmds(T_COMM,pline);
      }
    else
      ++plineptr;
    }
  ++plineptr;
  ++plineptr;
  }

p_number()                      /* Wort mit Ziffer am Anfang            */
/*------*/
  {
  do
    p_addch(*plineptr);
  while (alnum(*++plineptr));
  }

p_word()                        /* Wort mit Buchstaben am Anfang        */
/*----*/
  {
  char  c,name[NAMESIZE],*m;

  m=name;
  do
    if (m<name+NAMEEND)         /* Wortlnge begrenzen                  */
      *m++=*plineptr;
  while (alnum(*++plineptr));
  *m='\0';
  if ((m=mac_find(name))==NULL) /* auf Makronamen testen                */
    m=name;
  while (c=*m++)                /* Wort oder Ersetzung bernehmen       */
    p_addch(c);
  }

/* --- Kommandos ------------------------------------------------------ */

p_command()                     /* Kommando ausfhren                   */
/*-------*/
  {
  ilineptr=iline;
  *ilineptr='\0';
  if (p_match("define"))        /* "#define"                            */
    {
    p_expand();                 /* Makronamen expandieren               */
    p_addch(' ');               /* Leerzeichen als Trenner              */
    *ilineptr='\0';
    while (*plineptr)
      {
      p_expand();               /* Rest der Zeile expandieren           */
      if (ilineptr>=iline+ILINEEND)
        {
        error("Zeile zu lang");
        break;
        }
      }
    mac_add();                  /* Makro merken                         */
    }
  else if (p_match("include"))  /* "#include"                           */
    {
    p_expand();                 /* Dateinamen expandieren               */
    p_include();                /* Datei ffnen                         */
    }
  else if (p_match("asm"))      /* "#asm"                               */
    p_asm();
  else
    error("Unbekannter Preprozessor-Befehl");
  pline[0]='\0';                /* Befehlszeile lschen                 */
  }

p_include()                     /* Einfgedatei ffnen                  */
/*-------*/
  {
  if (in_i>=INMAXF-1)
    error("Zu viele \"#include\"-Dateien");
  else if (iline[0]!='"')
    error("Fehlendes Anfhrungszeichen");
  else
    {
    ilineptr=iline;
    while (*++ilineptr!='"');
    *ilineptr='\0';             /* Dateinamen isolieren                 */
    if (inopen(iline+1)==0)
      error("Kann \"include\"-Datei nicht ffnen");
    else if (verbose)
      {
      sprintf(oline,"#include \"%s\"\r\n",iline+1);
      puts(oline);
      }
    }
  }

p_asm()                         /* Direkte Ausgabe an Zieldatei         */
/*---*/
  {
  while (1)
    {
    if (ineof)
      {
      error("Fehlendes \"#endasm\"");
      break;
      }
    in_line();                  /* Zeile ohne bersetzung lesen         */
    for (plineptr=pline; isspace(*plineptr); ++plineptr);
    if (strncmp(plineptr,"#endasm",7)==0)
      break;
    moutstr(pline);             /* und ausgeben                         */
    mouteol();
    }
  moutcmds(T_COMM,"#endasm");
  }

p_match(char *s)                /* Wortvergleich                        */
/*------------*/
  {
  int   k;

  k=strlen(s);
  if (strncmp(plineptr,s,k))
    return(0);
  plineptr=plineptr+k;
  return(1);
  }

/* --- Makros verwalten ----------------------------------------------- */

mac_add()                       /* Makro hinzufgen                     */
/*-----*/
  {
  ilineptr=iline;
  if (alpha(*ilineptr))
    while (alnum(*++ilineptr));
  *ilineptr='\0';               /* Namen isolieren                      */
  if (ilineptr==iline)
    {
    error("Fehlerhafter Makroname");
    return;
    }
  ilineptr=iline;
  while (mac_put(*ilineptr++)); /* Namen kopieren                       */
  while (mac_put(*ilineptr++)); /* Krper kopieren                      */
  if (macptr>=mac+MACEND)
    error("Zuviele \"define\"s");
  }

mac_put(char c)                 /* Makrozeichen hinzufgen              */
/*-----------*/
  {
  if (macptr<mac+MACEND)
    *macptr++=c;
  return(c);
  }

mac_find(char *name)            /* Makro suchen                         */
/*----------------*/
  {
  char  *m,*n;

  m=mac;
  while (m<macptr)
    {
    n=name;
    while (*m==*n++)            /* Namen vergleichen                    */
      if (*m++=='\0')
        return(m);              /* bei Fund Zeiger auf Krper           */
    while (*m++);
    while (*m++);               /* nchstes Makro                       */
    }
  return(NULL);
  }

/* --- Zeichentests --------------------------------------------------- */

alpha(char c)                   /* Buchstabe oder Unterstrich           */
/*---------*/
  {
  return((c>='a' & c<='z') |
         (c>='A' & c<='Z') |
          c=='_' );
  }

alnum(char c)                   /* Buchstabe, Ziffer oder Unterstrich   */
/*---------*/
  {
  return((c>='a' & c<='z') |
         (c>='A' & c<='Z') |
         (c>='0' & c<='9') |
          c=='_' );
  }

/* --- Test auf Tokens ------------------------------------------------ */

matchx(char *s)                 /* Test auf beliebigen Text, wird bei   */
/*-----------*/                 /* Fund bersprungen                    */
  {
  int   k;

  ppactch();
  k=strlen(s);
  if (strncmp(ilineptr,s,k))    /* Text ungleich ?                      */
    return(0);
  ilineptr=ilineptr+k;
  return(1);
  }

matcha(char *s)                 /* Test auf Text ohne folgende "alnums" */
/*-----------*/                 /* wird bei Fund bersprungen           */
  {
  int   k;

  ppactch();
  k=strlen(s);
  if (strncmp(ilineptr,s,k))    /* Text ungleich ?                      */
    return(0);
  if (alnum(ilineptr[k]))       /* danach alphanumerisches Zeichen ?    */
    return(0);
  ilineptr=ilineptr+k;
  return(1);
  }

/* --- Fehlerhaften Text berspringen --------------------------------- */

skipjunk()                      /* aktuelle Stelle berspringen         */
/*------*/
  {
  if (alnum(ppgetch()))
    while (alnum(ppch0()))
      ppgetch();
  else
    while (alnum(ppch0())==0)
      {
      if (ppch0()=='\0')
        break;
      ppgetch();
      }
  }

/* === Ende =========================================================== */
